# Python module
""" cache.py -- manage documents cache

An index describes the contents of the cache.  Each entry of the index is made
of (id, time, type, url).  Entries are stored in .../cache/index as
id:time:type:url.
Internally we keep a map url->(id,type) to be able to find back some document
from its url, and a list of (id,time,type,url) sorted by id to be able to
purge the cache periodically.  Entries in the index cache file are stored in
this order for faster loading.

NB:  this assumes (and makes so) idX > idY <=> timeX > timeY
NB:  cache file removing is done by an idle callback, while purge checking is
done periodically.  Additionally index renaming is done when it becomes too
big.
"""

import string, os
import safe, settings, clock
from _gtk import gtk_idle_add, gtk_timeout_add

urlId = {}
nextId = 1
purgeByTime = []
changedSinceLastSave = None
dataFileFmt = "%04d"

cacheDir = settings.getUserCache ()

# --- getting and updating

class LoadFromCache :
    def __init__ (self, fp, url, contentType) :
        self.url = url
        self.contentType = contentType
        self.fileno = fp.fileno
        self.read = fp.read
        self.close = fp.close
    def info (self) : return self
    def geturl (self) : return self.url
    def gettype (self) : return self.contentType

def get (u) :
    if not urlId.has_key (u) :
        return None
    (i,t) = urlId[u]
    name = os.path.join (cacheDir, dataFileFmt % i)
    try :
        fp = open (name, "rb")
        f = LoadFromCache (fp, u, t)
    except IOError :
        f = None
    return f

class SaveToCache :
    def __init__ (self, fp, pos, entry) :
        self.fp = fp
        self.pos = pos
        self.entry = entry
        self.write = fp.write
    def close (self) :
        (i, t, contentType, url) = self.entry
        purgeByTime[self.pos] = self.entry
        urlId[url] = (i, contentType)
        self.pos = None
        global changedSinceLastSave
        changedSinceLastSave = 1
        gtk_timeout_add (settings.cacheRecordDelay*1000, save, ())
        return self.fp.close ()

def put (u, contentType) :
    if cacheDir is None :
        return None
    global nextId
    i = nextId
    name = os.path.join (cacheDir, dataFileFmt % i)
    try :
        fp = open (name, "wb")
    except IOError :
        return None
    nextId = nextId + 1
    s = int (clock.current)
    pos = len (purgeByTime)
    purgeByTime.append (None) # reserve space in list
    f = SaveToCache (fp, pos, (i, s, contentType, u))
    return f

# --- saving and loading

def save () :
    global changedSinceLastSave
    if not changedSinceLastSave :
        return
    if cacheDir is None :
        return
    x = lambda args: "%d:%d:%s:%s\n" % args
    l = map (x, filter (None, purgeByTime))
    s = string.join (l, "")
    file = os.path.join (cacheDir, "index")
    safe.save (file, s)
    changedSinceLastSave = None

def load () :
    if cacheDir is None :
        return
    file = os.path.join (cacheDir, "index")
    try :
        s = safe.load (file)
    except IOError :
        return

    l = string.split (s, "\n")
    # remove result of splitting trailing EOL
    if l[-1:] == [""] :
        del l[-1:]

    newUrlId = {}
    i = 0
    nextPurgeByTime = []
    for e in l :
        if e is None : continue # removed entry
        (i,s,t,u) = string.split (e, ":", 3)
        i = int (i)
        s = int (s)
        newUrlId[u] = (i, t)
        nextPurgeByTime.append ((i, s, t, u))

    global urlId, nextId, purgeByTime
    (urlId, nextId, purgeByTime) = (newUrlId, i+1, nextPurgeByTime)

load ()

